#!/usr/bin/env python3
"""Command line tool to configure TI CDCE(L)9XX clock generators across an I2C bus

   Requirements
   python (>=3.5)
   smbus (>=1.1.post2)
   Adafruit-PlatformDetect(>=3.19.4)
   cdce9xx (>=0.1)
"""

__version__ = '0.8.2'
__author__ = 'Fred Fierling'
__copyright__ = 'Copyright 2022 Spukhafte Systems Limited'

import os
import sys
import argparse

from math import modf
from smbus import SMBus

from spukhafte.cdce9xx import harmonic_in_band, CDCE9xx, PLL
from adafruit_platformdetect import Detector

# Error return values
SUCCESS = 0
FIN_ERROR = 1
FOUT_ERROR = 2
PDIV_ERROR = 3
SOLUTION_ERROR = 4
SET_ERROR = 5
VCO_ERROR = 6
PLL_ERROR = 7

LANGUAGE = 'english'  # TODO international support

# GPS band spec
# (center_freq, span) in Hz
GPS_BANDS = (
        (1567.74750e6, 1583.09250e6, 'GPS L1C/A'),  # Subset of E1B/C
#       (1563.42000e6, 1587.42000e6, 'QZSS'),
#       (1559.00000e6, 1591.00000e6, 'Galileo E1B/C'),
#       (1589.06250e6, 1605.37500e6, 'GLONASS L1OF'),
#       (1559.05200e6, 1591.78800e6, 'BeiDou B1'),
)

# Functions
def auto_int(x):
    """Convert a string into an integer.

    :param x: String to convert
    :return: value in x
    :rtype: int
    """
    return int(x, 0)

def primes(n):
    """Return a list of primes < n and > 2.

    :param n: Ceiling
    :type n: int
    :return: primes
    :rtype: list
    """

    from itertools import compress
    sieve = bytearray([True]) * (n//2)

    for i in range(3, int(n**0.5)+1, 2):
        if sieve[i//2]:
            sieve[i*i//2::i] = bytearray((n-i*i-1)//(2*i)+1)

    return [2, *compress(range(3, n, 2), sieve[1:])]

def factorize(n):
    """Return tuples of n's prime factors and their exponents.

    :param n: Number to factor
    :type n: int
    :return: tuples of prime factors and exponents
    :rtype: tuple of tuples
    """
    pf = []

    for p in primes(int(n**0.5)+1):
        if p*p > n:
            break
        count = 0
        while not n%p:
            n //= p
            count += 1
        if count > 0:
            pf.append((p, count))

    if n > 1:
        pf.append((n, 1))

    return pf

def print_factors(prefix, product):
    """Pretty-print factors of a product and their corresponding exponents.

    :param prefix: String prefixed to output
    :param product: List of (factor, exponent) to be printed
    :type product: float
    """

    f, i = modf(product)
    out = ''
    for b, e in factorize(int(i)):
        if out is '':
            out += prefix
        else:
            out += "*"

        if e == 1:
            out += "%d" % (b)
        else:
            out += "%d^%d" % (b, e)

    if f:
        out += '%+.4f' % f

    print(out, file=sys.stderr)

def main(device='_'):
    """Command line tool for configuring clock generators connected to local system on an I2C bus.
    """

    # Detect and set-up for clock generators in specific products
    vendor, model = (device.upper()).split('_')
    if model.startswith('K'):
        if model == 'K15' or model == 'K25':
            DEF_FIN = 30.72e6  # u-blox LEA-M8F REF_FREQ_OUT
            DEF_CONFIG = (
                ("INCLK", 2),  # LVCMOS
                ("M2", 0),  # R1: PDIV1
               #("M3", 0),  # R2: PDIV1
                ("PDIV2", 0), # Reset and stand-by
                ("PDIV3", 0), # Reset and stand-by
            )  # TODO check for R2 board

            EXCLUDE = GPS_BANDS
            N_PLLS = 1
           #DEF_PDIV_MAP = (1, 3)
           #VALID_PDIVS = (1, 2)  # R1: Y3 connects to a test point
            VALID_PLLS = (0, 1)
            DEF_PDIV_MAP = (1, 1)
            VALID_PDIVS = (1, 3)  # R2: Y2 connects to a test point

        else: 
            DEF_FIN = 27e6  # XTAL
            DEF_CONFIG = (
                    (("MUX1", 0),
                    ("Y1_ST0", 3),
                    ("XCSEL", 5)) +  # Trim oscillator R0
                    tuple(("PDIV%d" % i, 0) for i in range(1,8))  # Reset and stand-by
                    )

            EXCLUDE = ()
            N_PLLS = 3
           #DEF_PDIV_MAP = (1, 1, 4) # R1
           #VALID_PDIVS = (1, 2, 4)  # R1: Y3, Y5 connect to test points
            VALID_PLLS = (0, 1, 2, 3)  # Zero means null pll
            DEF_PDIV_MAP = (1, 1, 5, 7)
            VALID_PDIVS = (1, 3, 4, 5, 7)  # R2: Y2, Y6 connect to test points
    else:
        DEF_FIN = 27e6  # XTAL
        DEF_CONFIG = (
            ("PDIV1", 0), # Reset and stand-by
            ("PDIV2", 0), # Reset and stand-by
            ("PDIV3", 0), # Reset and stand-by
        )  #

        EXCLUDE = ()
        N_PLLS = 1
        VALID_PLLS = (0, 1, 2)
        DEF_PDIV_MAP = (1, 1, 1)
        VALID_PDIVS = (1, 2, 3)  # R2: Y2 connects to a test point

    # Detect system hardware and software
    HW = Detector()

    if HW.board.id == 'BEAGLEBONE_BLACK':
        VALID_BUSES = (0,1,2)
        DEF_BUS = 1 if model == 'K25' else 2
    elif HW.board.id.startswith('RASPBERRY_PI'):
        VALID_BUSES = (1,2)
        DEF_BUS = 1
    elif HW.board.id == 'BEAGLEBONE_AI':
        VALID_BUSES = (0,3)
        DEF_BUS = 3
    else:
        VALID_BUSES = (0,1,2)
        DEF_BUS = 0

    ENV_PREFIX = vendor + '_' + model

    var = ENV_PREFIX + '_BUS'
    if var in os.environ:
        DEF_FIN = auto_int(os.environ[var])

    var = ENV_PREFIX + '_ADDR'
    if var in os.environ:
        DEF_ADDR = auto_int(os.environ[var])
    else:
        DEF_ADDR = CDCE9xx.def_addr(N_PLLS)

    var = ENV_PREFIX + '_FIN'
    if var in os.environ:
        DEF_FIN = float(os.environ[var])

    # Parse arguments
    parser = argparse.ArgumentParser(
        description=("configure %s's PLL" % model),
        epilog='no arguments: list PLL1 configuration')

    group = parser.add_mutually_exclusive_group()

    # Mutually exclusive arguments
    group.add_argument('fout', type=float, nargs='?',
        help="output frequency (Hz)",
        metavar='FOUT'
    )

    group.add_argument('-nmp', type=int, nargs=3,
        metavar=('N', 'M', 'PDIV'),
        help="where FOUT = N/M * FIN/PDIV"
    )

    # Optional arguments
    parser.add_argument('-a', dest='addr', type=auto_int, default=DEF_ADDR,
        help="I2C address of CDCE9%d%d (default=0x%x)"
              % (N_PLLS, (1 + 2 * N_PLLS), DEF_ADDR))

    parser.add_argument('-b', dest='bus', type=int, choices=VALID_BUSES,
        default=DEF_BUS,
        help="I2C bus number (default=%d)" % (DEF_BUS))

    parser.add_argument('-d', type=int, default=None,
        choices=VALID_PDIVS,
       #metavar=('PDIV'),
        help="target PDIV index")

    parser.add_argument('-fin', dest='fin', type=float, default=DEF_FIN,
        help="input frequency Xin/CLK, (default=%.2f MHz)" % (DEF_FIN/1e6))

    parser.add_argument('-g', action="store_true", 
        help="allow VCO harmonics in GPS bands")

    parser.add_argument('-i', type=argparse.FileType('r'),
        metavar=('FILE'),
        help='file containing config JSON')

    parser.add_argument('-j', action="store_true", 
        help="output configuration in JSON")

    parser.add_argument('-k', dest='pdiv10', action="store_true", default=False,
        help="solve for a ten bit divider")

    parser.add_argument('-p', type=int, default=CDCE9xx.DEF_PLL,
        choices=VALID_PLLS,
       #metavar=('PLL'),
        help="target PLL index (default=%d, 0: null device)" % CDCE9xx.DEF_PLL)

    parser.add_argument('-r', action="store_true", 
        help="reset to factory default")

    parser.add_argument('-s', type=int, default=0, choices=(0, 1),
        help="target PLL state, (default=%(default)s)")

    parser.add_argument('-w', action="store_true", help="write EEPROM")
    parser.add_argument('-v', default=0, action="count", help="verbosity level")
    parser.add_argument('-z', action="store_true",
        help="dump zero value config JSON")
 
    args = parser.parse_args()
 
    # Check fin
    if not CDCE9xx.fin_valid(args.fin):
        parser.error('fin out of range')
        return(FIN_ERROR)

    # Check fout
    if args.fout and not PLL.fout_valid(args.fout, args.pdiv10):
        parser.error('fout out of range')
        return(FOUT_ERROR)

    # Set target divider of PLL
    if args.d is None:
        args.d = DEF_PDIV_MAP[args.p]

    # Check size of divider
    if args.pdiv10:
        if args.d in (0, 1):
            max_pdiv = PLL.MAX_PDIV1
        else:
            parser.error('-10b and -d=%d not allowed' % args.d)
            return(PDIV_ERROR)
    else:
        max_pdiv = PLL.MAX_PDIV

    if args.v:
        print('HW=%s' % HW.board.id)
        print_factors('FIN=', args.fin)
        
        if args.fout:
            print_factors('FOUT=', args.fout)  

    if args.g:
        EXCLUDE = ()

    # if not args.n:
    i2c = SMBus(args.bus)
    i2c.open(args.bus)

    clock_gen = CDCE9xx(N_PLLS, i2c, args.addr)

    if args.r:
        # Set all fields to factory defaults
        clock_gen.factory_default()
        clock_gen.set_fields(DEF_CONFIG)

    if args.fout or args.nmp:
        if args.fout:
            n, m, pdiv, fvco, error = PLL.find_n_m_pdiv(args.fin, args.fout, max_pdiv,
                                              exclude=EXCLUDE, debug=args.v)
            if None in (n, m, pdiv):
                parser.error('solution not found for FIN=%.2e FOUT=%.2e' 
                        % (args.fin, args.fout))
                return(SOLUTION_ERROR)

        elif args.nmp:
            if PLL.nmp_valid(args.nmp, args.pdiv10):
                n, m, pdiv = args.nmp
            else:
                parser.error('N, M, or PDIV out of range')
                return(SET_ERROR)

            fvco = (args.fin * n) / m
            band = harmonic_in_band(fvco, EXCLUDE)

            if band is not None:
                parser.error('harmonic of FVCO=%.2e in: %s' % (fvco, band))

        vco = PLL.vco_range(fvco)
        if vco is None:
            print('%s: FVCO=%.2e out of range' % (__name__, fvco), file=sys.stderr)
            return(VCO_ERROR)

        if clock_gen.set_pll_pdiv(args.p, n, m, vco, args.d, pdiv,
                                state_n=args.s, debug=args.v):
            if args.v:
                format = 'N=%04d M=%03d PDIV=%04d FVCO=%.2e '

                # Distinguish perfect solution
                format += 'ERROR=%f' if error else 'ERROR=%d'
                print(format % (n, m, pdiv, fvco, error), file=sys.stderr)
        else:
            print('%s: error: set failed' % (__name__), file=sys.stderr)
            return(SET_ERROR)
    else:
        if args.p:
            n, m, pdiv = clock_gen.get_nmp(args.p, args.d, args.s)
        else:
            parser.error("can't read PLL0")
            return(PLL_ERROR)

    fvco, fout = PLL.calculate_fvco_fout(args.fin, n, m, pdiv)

    if fvco is not None:
        fmt = ('FIN=%.3e PLL%d_%d: N=%04d M=%03d PDIV%d=%04d FVCO=%.3e ' +
               ('FOUT=%d' if fout.is_integer() else 'FOUT=%.3e')
              )

        print(fmt % (args.fin, args.p, args.s, n, m, args.d, pdiv, fvco, fout),
              file=sys.stderr)

    if not sys.stdin.isatty():
        clock_gen.load(sys.stdin)

    # Configure PLL with JSON
    if args.i and args.i.readable():
        # Configure PLL using JSON on stdin
        clock_gen.load(args.i)

    if args.j:
        clock_gen.dump(args.z)

    if args.w:
        i = clock_gen.write_eeprom()
        
        if i:
            if args.v:
                print('EEPIPs=%d' % (i), file=sys.stderr)
        else:
            print('EEPROM write failed', file=sys.stderr)

    i2c.close()
 
    return(SUCCESS)

def k15():
    """Configure Spukhafte K15 on MIKROE-1857"""
    main(device='ssl_k15')

def k16():
    """Configure Spukhafte K16 on MIKROE-1857"""
    main(device='ssl_k16')

def k25():
    """Configure Spukhafte K25 on SSL functional test interface"""
    main(device='ssl_k25')

if __name__ == "__main__":
    main()



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='spukhafte-cdce9xx',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/guides/single-sourcing-package-version/
    version='0.8.2',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Python3 I2C Driver & Application for TI CDCE9XX Clock Generators',  # Optional

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # Denotes that our long_description is in Markdown; valid values are
    # text/plain, text/x-rst, and text/markdown
    #
    # Optional if long_description is written in reStructuredText (rst) but
    # required for plain-text or Markdown; if unspecified, "applications should
    # attempt to render [the long_description] as text/x-rst; charset=UTF-8 and
    # fall back to text/plain if it is not valid rst" (see link below)
    #
    # This field corresponds to the "Description-Content-Type" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type='text/markdown',  # Optional (see note above)

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/spukhafte-sys/cdce9xx',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Spukhafte Systems Limited',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='fff@spukhafte.net',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Programming Language :: Python :: 3.10",
        'Programming Language :: Python :: 3 :: Only',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a list of additional keywords, separated
    # by commas, to be used to assist searching for the distribution in a
    # larger catalog.
    keywords='i2c, cdce913, cdcel913, cdce925, cdcel925, cdce937, cdcel937, phase lock loop',  # Optional

    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    #package_dir={'': 'src'},  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['dist', 'docs', 'tests']),  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. See
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.7.3, <4',  # Tested on 3.7.3

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/discussions/install-requires-vs-requirements/
    install_requires=['smbus>=1.1', 'Adafruit-PlatformDetect>=3.19.4'],  # Minimum tested on BBB

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    #extras_require={  # Optional
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #package_data={  # Optional
    #    'lib': ['package_data.dat'],
    #},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
        'console_scripts': [
            'cdce9xx=spukhafte.bin.config:main',
            'k15_config=spukhafte.bin.config:k15',
            'k16_config=spukhafte.bin.config:k16',
            'k25_config=spukhafte.bin.config:k25',
        ],
    },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/spukhafte-sys/cdce9xx/issues',
        'Source': 'https://github.com/spukhafte-sys/cdce9xx/',
    },
)
